/**
 *
 * $Id: GPForceModels.java 616 2017-10-13 10:18:42Z ffsm $
 *
 * HISTORIQUE
* VERSION:2.1:DM:DM-630:11/10/2019:[GENOPUS] Modifications widgets pour FT 336 de PSIMU
* VERSION:2.1:DM:DM-635:11/10/2019:[GENOPUS] Montee de niveau GENIUS 1.9
* VERSION:2.1:DM:DM-617:11/10/2019:[GENOPUS] Widget integrateur numerique
 * 
 * VERSION:1.2:DM:DM-33:22/09/2016:Class creation.
 * VERSION:1.2:DM:DM-136:07/11/2016:Added license.
 * VERSION:1.2.1:DM:DM-159:31/01/2017: Implementing GPPatrius interface.
 * VERSION:1.2.1:FA:FA-144:24/02/2017:Updated constraints.
 * VERSION:1.3:DM:DM-165:04/07/2017:PATRIUS upgrade to V3.3
 * VERSION:1.3:DM:DM-192:21/07/2017:Initialize the attraction model by constructor.
 * VERSION:1.3:DM:DM-193:27/09/2017:Copy/paste menu active by default, implementing GCondensedIsModifiedInterface.
 * VERSION:2.0:DM:DM-213:31/05/2018:Upgrade to PATRIUS 4.1 and remove GENOPUS Custom classes
 * VERSION:2.0:DM:DM-218:13/09/2018:Upgrade to GENIUS 1.8
 * 
 * FIN-HISTORIQUE
 *
 * This software is part of GENOPUS, 
 * A CNES Library of Flight Dynamics MMI in Java 
 * 
 * Copyright 2016-2026 CNES Centre National d’Etudes Spatiales 
 * Licensed to Centre National d’Etudes Spatiales (CNES) under one or more 
 * contributor license agreements. See the NOTICE file distributed with 
 * this work for additional information regarding copyright ownership. 
 * CNES licenses this file to You under the Apache License, Version 2.0 
 * (the "License"); you may not use this file except in compliance with 
 * the License. You may obtain a copy of the License at 
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.  
 * 
 **/

package observer;

import java.util.HashMap;

import fr.cnes.genius.constraints.GConstraint;
import fr.cnes.genius.exception.GArithmeticException;
import fr.cnes.genius.exception.GComponentException;
import fr.cnes.genius.exception.GException;
import fr.cnes.genius.exception.GFileManipulatorException;
import fr.cnes.genius.exception.GRuntimeException;
import fr.cnes.genius.highLevel.GChoice;
import fr.cnes.genius.highLevel.GLabelWithIndicator;
import fr.cnes.genius.interval.GIntervalAbstract.GStatus;
import fr.cnes.genius.logger.GLogger.Type;
import fr.cnes.genius.lowLevel.GLabel;
import fr.cnes.genius.lowLevel.GPanel;
import fr.cnes.genius.lowLevel.GSeparator;
import fr.cnes.genius.main.GClear;
import fr.cnes.genius.main.GComponent;
import fr.cnes.genius.main.GCondensedIsModified;
import fr.cnes.genius.main.GCondensedIsModifiedInterface;
import fr.cnes.genius.main.GCondensedStatus;
import fr.cnes.genius.main.GCondensedStatusInterface;
import fr.cnes.genius.main.GEnvironment;
import fr.cnes.genius.main.GEvent;
import fr.cnes.genius.main.GListener;
import fr.cnes.genius.main.GReadWrite;
import fr.cnes.genius.observer.GObservable;
import fr.cnes.genius.observer.GObserver;
import fr.cnes.genius.observer.GObserverList;
import fr.cnes.genopus.bodies.GPOneAxisEllipsoid;
import fr.cnes.genopus.exception.GPEventException;
import fr.cnes.genopus.exception.GPForcesException;
import fr.cnes.genopus.exception.GPFrameException;
import fr.cnes.genopus.forces.GPAeroForceModel;
import fr.cnes.genopus.forces.GPAeroForceModel.AtmosphereType;
import fr.cnes.genopus.forces.GPEarthPotentialForceModel;
import fr.cnes.genopus.forces.GPEarthPotentialForceModel.AttractionModelsEnum;
import fr.cnes.genopus.forces.GPOceanTidesForceModel;
import fr.cnes.genopus.forces.GPRediffusedSRPForceModel;
import fr.cnes.genopus.forces.GPSRPForceModel;
import fr.cnes.genopus.forces.GPTerrestrialTidesForceModel;
import fr.cnes.genopus.forces.GPThirdBodyForceModel;
import fr.cnes.genopus.main.GPPatriusWithObjects;
import fr.cnes.sirius.patrius.assembly.Assembly;
import fr.cnes.sirius.patrius.bodies.CelestialBody;
import fr.cnes.sirius.patrius.bodies.CelestialBodyFactory;
import fr.cnes.sirius.patrius.bodies.ExtendedOneAxisEllipsoid;
import fr.cnes.sirius.patrius.bodies.MeeusMoon;
import fr.cnes.sirius.patrius.bodies.MeeusSun;
import fr.cnes.sirius.patrius.bodies.MeeusSun.MODEL;
import fr.cnes.sirius.patrius.forces.ForceModel;
import fr.cnes.sirius.patrius.forces.ForceModelsData;
import fr.cnes.sirius.patrius.forces.atmospheres.Atmosphere;
import fr.cnes.sirius.patrius.forces.drag.DragForce;
import fr.cnes.sirius.patrius.forces.gravity.ThirdBodyAttraction;
import fr.cnes.sirius.patrius.forces.gravity.tides.OceanTides;
import fr.cnes.sirius.patrius.forces.gravity.tides.TerrestrialTides;
import fr.cnes.sirius.patrius.forces.radiation.RediffusedRadiationPressure;
import fr.cnes.sirius.patrius.forces.radiation.SolarRadiationPressureEllipsoid;
import fr.cnes.sirius.patrius.frames.Frame;
import fr.cnes.sirius.patrius.stela.bodies.MeeusMoonStela;
import fr.cnes.sirius.patrius.utils.exception.PatriusException;

/**
 * Class containing force models definition widget.
 * 
 * @author ffsm
 * 
 * @genius.defaultConstraints 0:GPEarthPotentialForceModel:earthPotentialForceModel:"newline"<br>
 *                            (see table below);
 *                            1:GPThirdBodyForceModel:thirdBodyForceModel:"newline"<br>
 *                            (see table below); 2:GPAeroForceModel:aeroForceModel:"newline"<br>
 *                            (see table below); 3:GPSRPForceModel:srpForceModel:"newline"<br>
 *                            (see table below);
 *                            4:GPRediffusedSRPForceModel:rediffusedSRPForceModel:"newline"<br>
 *                            (see table below);
 *                            5:GPOceanTidesForceModel:oceanTidesForceModel:"newline"<br>
 *                            (see table below)
 * @genius.defaultConstraints [[<i>earthPotentialForceModel</i>]]0,
 *                            0:GLabelWithIndicator:earthPotentialLabel:0, 0, 0 =>
 *                            "newline , gapleft 0 , split 2"<br>
 *                            0, 0, 1 => "wmin 128"; 0, 1:GChoice:attractionModelChoice:0, 1, 0, 0
 *                            => "newline , gapleft 20 , split 2"<br>
 *                            0, 1, 0, 1 => (null)<br>
 *                            0, 1, 1 => "split"; 0,
 *                            2:GComboBoxWithLabel:earthPotentialFileComboBox:0, 2, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            0, 2, 0, 1 => (null)<br>
 *                            0, 2, 1 => (null); 0, 3:GEntryInt:maxOrdDegEntryInt:0, 3, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            0, 3, 0, 1 => (null)<br>
 *                            0, 3, 1 => (null); 0, 4:GEntryInt:degreeEntryInt:0, 4, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            0, 4, 0, 1 => (null)<br>
 *                            0, 4, 1 => (null); 0, 5:GEntryInt:orderEntryInt:0, 5, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            0, 5, 0, 1 => (null)<br>
 *                            0, 5, 1 => (null); 0, 6:GEntryInt:degreeOptionalEntryInt:0, 6, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            0, 6, 0, 1 => (null)<br>
 *                            0, 6, 1 => (null); 0, 7:GEntryInt:orderOptionalEntryInt:0, 7, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            0, 7, 0, 1 => (null)<br>
 *                            0, 7, 1 => (null); 0, 8:GChoice:coeffsComputedOnceChoice:0, 8, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            0, 8, 0, 1 => (null)<br>
 *                            0, 8, 1 => "split"
 * @genius.defaultConstraints [[<i>thirdBodyForceModel</i>]]1,
 *                            0:GCheckBoxWithIndicator:thirdBodyCheckBox:1, 0, 0, 0 =>
 *                            "newline , gapleft 0 , split 2"<br>
 *                            1, 0, 0, 1 => (null)<br>
 *                            1, 0, 1 => (null); 1,
 *                            1:GCheckBoxWithIndicator:thirdBodyMoonCheckBox:1, 1, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            1, 1, 0, 1 => (null)<br>
 *                            1, 1, 1 => (null); 1, 2:GCheckBoxWithIndicator:thirdBodySunCheckBox:1,
 *                            2, 0, 0 => "newline , gapleft 20 , split 2"<br>
 *                            1, 2, 0, 1 => (null)<br>
 *                            1, 2, 1 => (null); 1,
 *                            3:GCheckBoxWithIndicator:thirdBodyVenusCheckBox:1, 3, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            1, 3, 0, 1 => (null)<br>
 *                            1, 3, 1 => (null); 1,
 *                            4:GCheckBoxWithIndicator:thirdBodyMarsCheckBox:1, 4, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            1, 4, 0, 1 => (null)<br>
 *                            1, 4, 1 => (null); 1,
 *                            5:GCheckBoxWithIndicator:thirdBodyJupiterCheckBox:1, 5, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            1, 5, 0, 1 => (null)<br>
 *                            1, 5, 1 => (null)
 * @genius.defaultConstraints [[<i>aeroForceModel</i>]]2,
 *                            0:GCheckBoxWithIndicator:aeroForceCheckBox:2, 0, 0, 0 =>
 *                            "newline , gapleft 0 , split 2"<br>
 *                            2, 0, 0, 1 => (null)<br>
 *                            2, 0, 1 => (null); 2, 1:GChoice:aeroModelChoice:2, 1, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            2, 1, 0, 1 => (null)<br>
 *                            2, 1, 1 => "split"; 2, 2:GChoice:solarActivityChoice:2, 2, 0, 0 =>
 *                            "newline , gapleft 40 , split 2"<br>
 *                            2, 2, 0, 1 => (null)<br>
 *                            2, 2, 1 => "split"; 2, 3:GEntryReal:f107EntryReal:2, 3, 0, 0 =>
 *                            "newline , gapleft 60 , split 2"<br>
 *                            2, 3, 0, 1 => (null)<br>
 *                            2, 3, 1 => (null)<br>
 *                            ; 2, 4:GEntryReal:apEntryReal:2, 4, 0, 0 =>
 *                            "newline , gapleft 60 , split 2"<br>
 *                            2, 4, 0, 1 => (null)<br>
 *                            2, 4, 1 => (null)<br>
 *                            ; 2, 5:GButton:loadSolarActivityFileButton:"newline , skip 1"; 2,
 *                            6:GEntryString:loadedSolarActivityFileEntryString:2, 6, 0, 0 =>
 *                            "newline , gapleft 60 , split 2"<br>
 *                            2, 6, 0, 1 => (null)<br>
 *                            2, 6, 1 => (null); 2, 7:GEntryReal:densityH0EntryReal:2, 7, 0, 0 =>
 *                            "newline , gapleft 40 , split 2"<br>
 *                            2, 7, 0, 1 => (null)<br>
 *                            2, 7, 1 => "split 2"<br>
 *                            2, 7, 2 => (null); 2, 8:GEntryReal:h0EntryReal:2, 8, 0, 0 =>
 *                            "newline , gapleft 40 , split 2"<br>
 *                            2, 8, 0, 1 => (null)<br>
 *                            2, 8, 1 => "split 2"<br>
 *                            2, 8, 2 => (null); 2, 9:GEntryReal:hScaleEntryReal:2, 9, 0, 0 =>
 *                            "newline , gapleft 40 , split 2"<br>
 *                            2, 9, 0, 1 => (null)<br>
 *                            2, 9, 1 => "split 2"<br>
 *                            2, 9, 2 => (null); 2, 10:GEntryReal:multipAeroForceEntryReal:2, 10, 0,
 *                            0 => "newline , gapleft 20 , split 2"<br>
 *                            2, 10, 0, 1 => (null)<br>
 *                            2, 10, 1 => (null)<br>
 * @genius.defaultConstraints [[<i>srpForceModel</i>]]3, 0:GCheckBoxWithIndicator:sRPCheckBox:3, 0,
 *                            0, 0 => "newline , gapleft 0 , split 2"<br>
 *                            3, 0, 0, 1 => (null)<br>
 *                            3, 0, 1 => (null); 3, 1:GEntryReal:refDistEntryReal:3, 1, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            3, 1, 0, 1 => (null)<br>
 *                            3, 1, 1 => "split 2"<br>
 *                            3, 1, 2 => (null); 3, 2:GEntryReal:refPressureEntryReal:3, 2, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            3, 2, 0, 1 => (null)<br>
 *                            3, 2, 1 => "split 2"<br>
 *                            3, 2, 2 => (null); 3, 3:GEntryReal:multipSRPEntryReal:3, 3, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            3, 3, 0, 1 => (null)<br>
 *                            3, 3, 1 => (null)<br>
 * @genius.defaultConstraints [[<i>rediffusedSRPForceModel</i>]]4,
 *                            0:GCheckBoxWithIndicator:rediffusedSRPCheckBox:4, 0, 0, 0 =>
 *                            "newline , gapleft 0 , split 2"<br>
 *                            4, 0, 0, 1 => (null)<br>
 *                            4, 0, 1 => (null); 4,
 *                            1:GCheckBoxWithIndicator:rediffusedAlbedoCheckBox:4, 1, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            4, 1, 0, 1 => (null)<br>
 *                            4, 1, 1 => (null); 4,
 *                            2:GCheckBoxWithIndicator:rediffusedInfraredCheckBox:4, 2, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            4, 2, 0, 1 => (null)<br>
 *                            4, 2, 1 => (null); 4, 3:GEntryReal:multipRediffusedAlbedo:4, 3, 0, 0
 *                            => "newline , gapleft 40 , split 2"<br>
 *                            4, 3, 0, 1 => (null)<br>
 *                            4, 3, 1 => (null)<br>
 *                            ; 4, 4:GEntryReal:multipRediffusedInfrared:4, 4, 0, 0 =>
 *                            "newline , gapleft 40 , split 2"<br>
 *                            4, 4, 0, 1 => (null)<br>
 *                            4, 4, 1 => (null)<br>
 * @genius.defaultConstraints [[<i>oceanTidesForceModel</i>]]5,
 *                            0:GCheckBoxWithIndicator:oceanTidesCheckBox:5, 0, 0, 0 =>
 *                            "newline , gapleft 0 , split 2"<br>
 *                            5, 0, 0, 1 => (null)<br>
 *                            5, 0, 1 => (null); 5, 1:GEntryReal:waterDensityEntryReal:5, 1, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            5, 1, 0, 1 => (null)<br>
 *                            5, 1, 1 => "split 2"<br>
 *                            5, 1, 2 => (null); 5, 2:GEntryInt:orderOceanTidesEntryInt:5, 2, 0, 0
 *                            => "newline , gapleft 20 , split 2"<br>
 *                            5, 2, 0, 1 => (null)<br>
 *                            5, 2, 1 => (null); 5, 3:GEntryInt:degreeOceanTidesEntryInt:5, 3, 0, 0
 *                            => "newline , gapleft 20 , split 2"<br>
 *                            5, 3, 0, 1 => (null)<br>
 *                            5, 3, 1 => (null); 5,
 *                            4:GCheckBoxWithIndicator:withAdmittanceCheckBox:5, 4, 0, 0 =>
 *                            "newline , gapleft 20 , split 2"<br>
 *                            5, 4, 0, 1 => (null)<br>
 *                            5, 4, 1 => (null)
 * 
 */
public class GPForceModelsLocal extends GPanel implements GReadWrite, GClear, GCondensedStatusInterface,
        GCondensedIsModifiedInterface, GPPatriusWithObjects<ForceModelsData>, GListener, GObservable {

    /** Global label. */
    private final GLabelWithIndicator m_label;

    /** Earth Potential widget. */
    private final GPEarthPotentialForceModel m_earthPotentialForceModel;

    /** Third body force choice IsVisible. */
    private final boolean m_thirdBodyForceIsVisible;
    /** Third Body widget. */
    private GPThirdBodyForceModel m_thirdBodyForceModel;

    /** Aerodynamic force IsVisible. */
    private final boolean m_aeroForceIsVisible;
    /** Aerodynamic force model. */
    private GPAeroForceModel m_aeroForceModel;

    /** Solar Radiation Pressure force IsVisible. */
    private final boolean m_srpForceIsVisible;
    /** Solar Radiation Pressure force model. */
    private GPSRPForceModel m_srpForceModel;

    /** Rediffused SRP force IsVisible. */
    private final boolean m_rediffusedSRPForceIsVisible;
    /** Rediffused SRP force model. */
    private GPRediffusedSRPForceModel m_rediffusedSRPForceModel;

    /** Ocean tides force IsVisible. */
    private final boolean m_oceanTidesForceIsVisible;
    /** Ocean tides force model. */
    private GPOceanTidesForceModel m_oceanTidesForceModel;

    /** Terrestrial tides force IsVisible. */
    private final boolean m_terrestrialTidesForceIsVisible;
    /** Terrestrial tides force model. */
    private GPTerrestrialTidesForceModel m_terrestrialTidesForceModel;

    /** Choice for ephemerides type. */
    private final GChoice m_ephemerisTypeChoice;

    /** Boolean to set Eart's ellipsoid widget visibility. */
    private boolean m_bodyGPOneAxisEllipsoidIsVisible = true;
    private final boolean m_earthGPOneAxisEllipsoidIsVisibleByDefault = true;
    /** Earth's ellipsoid shape. */
    private final GPOneAxisEllipsoid m_bodyGPOneAxisEllipsoid;

    /** Separators. */
    private final GSeparator[] m_separatorsArray = new GSeparator[7];

    /** Sun celestial bodies for the three types of ephemeris. */
    private final HashMap<TypeEphemeris, CelestialBody> m_sunBodiesMap;

    //JFG DEB MOD
    /** List of observers */
    private final GObserverList obsList;
//    /** Observer for GPVehicle. */
//    private GObserver observerGPVehicle;
    //JFG END MOD
    
    /** Enum containing ephemeris types. */
    public static enum TypeEphemeris {

        /** JPL ephemerides (DE405/DE406). */
        JPL("JPL"),

        /** Meeus ephemerides. */
        MEEUS("Meeus"),

        /** Meeus STELA ephemerides. */
        MEEUSSTELA("Meeus Stela");

        /** Ephemeris type label. */
        private String m_label;

        /**
         * Private constructor.
         * 
         * @param label ephemeris type label
         */
        private TypeEphemeris(final String label) {
            m_label = label;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            return m_label;
        }
    }

    /**
     * Empty constructor.
     */
    public GPForceModelsLocal() {
        this(null);
    }

    /**
     * Creates a new instance where all forces appear. There is a widget to define the Earth's
     * shape. The widgetId will be assigned automatically.
     * 
     * @param label text to display in the label.
     */
    public GPForceModelsLocal(final String label) {
        this(null, label, AttractionModelsEnum.CUNNINGHAM, null, true, true, true, true, true, true);
    }

    /**
     * Creates a new instance where all forces appear. There is a widget to define the Earth's
     * shape.
     * 
     * @param widgetId widget identifier, if null an identifier will be automatically created.
     * @param label text to display in the label.
     */
    public GPForceModelsLocal(final String widgetId, final String label) {
        this(widgetId, label, AttractionModelsEnum.CUNNINGHAM, null, true, true, true, true, true,
                true);
    }

    /**
     * Creates a new instance where the shown forces are customized. There is a widget to define the
     * Earth's shape. The widgetId will be assigned automatically.
     * 
     * @param label text to display in the label.
     * @param initialAttractionModel initial attraction model for the Earth potential.
     * @param availableAttractionModels available attraction models for the Earth potential.
     * @param thirdBodyForceIsVisible if true, third body force IsVisible
     * @param aeroForceIsVisible if true, aerodynamic force IsVisible
     * @param srPForceIsVisible if true, SRP force IsVisible
     * @param rediffusedSRPForceIsVisible if true, rediffused SRP force IsVisible
     * @param oceanTidesForceIsVisible if true, ocean tides force IsVisible
     * @param terrestrialTidesForceIsVisible if true, terrestrial tides force IsVisible
     */
    public GPForceModelsLocal(final String label, final AttractionModelsEnum initialAttractionModel,
            final AttractionModelsEnum[] availableAttractionModels, final boolean thirdBodyForceIsVisible,
            final boolean aeroForceIsVisible, final boolean srPForceIsVisible,
            final boolean rediffusedSRPForceIsVisible, final boolean oceanTidesForceIsVisible,
            final boolean terrestrialTidesForceIsVisible) {
        this(null, label, initialAttractionModel, availableAttractionModels,
                thirdBodyForceIsVisible, aeroForceIsVisible, srPForceIsVisible,
                rediffusedSRPForceIsVisible, oceanTidesForceIsVisible,
                terrestrialTidesForceIsVisible);
    }

    /**
     * Creates a new instance where the shown forces are customized. A predefined Earth's shape is
     * used. If null, a widget to define it will be available.
     * 
     * @param widgetId widget identifier, if null an identifier will be automatically created.
     * @param label text to display in the label.
     * @param initialAttractionModel initial attraction model for the Earth potential.
     * @param availableAttractionModels available attraction models for the Earth potential.
     * @param thirdBodyForceIsVisible if true, third body force IsVisible
     * @param aeroForceIsVisible if true, aerodynamic force IsVisible
     * @param srPForceIsVisible if true, SRP force IsVisible
     * @param rediffusedSRPForceIsVisible if true, rediffused SRP force IsVisible
     * @param oceanTidesForceIsVisible if true, ocean tides force IsVisible
     * @param terrestrialTidesForceIsVisible if true, terrestrial tides force IsVisible
     */
    public GPForceModelsLocal(final String widgetId, final String label,
            final AttractionModelsEnum initialAttractionModel,
            final AttractionModelsEnum[] availableAttractionModels, final boolean thirdBodyForceIsVisible,
            final boolean aeroForceIsVisible, final boolean srPForceIsVisible,
            final boolean rediffusedSRPForceIsVisible, final boolean oceanTidesForceIsVisible,
            final boolean terrestrialTidesForceIsVisible) {
        super(widgetId);

        // Generating label
        setNameAndCommentInConfigFile(label);
        m_label = new GLabelWithIndicator(composeWidgetId("label"), label);

        // Ephemerides type (it is needed to create it before the aero force model)
        m_ephemerisTypeChoice = new GChoice(composeWidgetId("ephemerisTypeChoice"),
                GEnvironment.tr("Ephemeris type"), TypeEphemeris.JPL);
        m_ephemerisTypeChoice.setNameInConfigFile("ephemeridesType");
        m_ephemerisTypeChoice.setCommentInConfigFile(null);

        // Earth's Shape (it is needed to create it before the aero force model)
        m_bodyGPOneAxisEllipsoid = new GPOneAxisEllipsoid(
                composeWidgetId("earthGPOneAxisEllipsoid"), GEnvironment.tr("Earth's Ellipsoid"));

        // Creating Sun Bodies
        m_sunBodiesMap = new HashMap<GPForceModelsLocal.TypeEphemeris, CelestialBody>(3);
        try {
            // For JPL ephemeris
            m_sunBodiesMap.put(TypeEphemeris.JPL, CelestialBodyFactory.getSun());
            // For Meeus ephemeris
            m_sunBodiesMap.put(TypeEphemeris.MEEUS, new MeeusSun());
            // For Meeus Stela ephemeris
            m_sunBodiesMap.put(TypeEphemeris.MEEUSSTELA, new MeeusSun(MODEL.STELA));
        } catch (final PatriusException e) {
            throw new GRuntimeException(e);
        } catch (final ArithmeticException e) {
            // if Sun body cannot be built.
            throw new GRuntimeException(e);
        }

        // Init earth potential subwidget
        m_earthPotentialForceModel = new GPEarthPotentialForceModel(
                composeWidgetId("earthPotentialForceModel"), GEnvironment.tr("Earth Potential"),
                initialAttractionModel, availableAttractionModels);

        // Init third subwidget
        m_thirdBodyForceIsVisible = thirdBodyForceIsVisible;
        if (m_thirdBodyForceIsVisible) {
            m_thirdBodyForceModel = new GPThirdBodyForceModel(
                    composeWidgetId("thirdBodyForceModel"), GEnvironment.tr("Third Body"));
        }

        // Init aero force subwidget
        m_aeroForceIsVisible = aeroForceIsVisible;
        if (m_aeroForceIsVisible) {
            m_aeroForceModel = new GPAeroForceModel(composeWidgetId("aeroForceModel"),
                    GEnvironment.tr("Atmospheric Force"));
            try {
                m_aeroForceModel.setBodyShape(getBodyShape());
            } catch (final GPForcesException e) {
                throw new GRuntimeException(e);
            } catch (final GArithmeticException e) {
                // If earth shape contains NaN (impossible comming from widget)
                throw new GRuntimeException(e);
            }
            m_aeroForceModel.setSun(getSun());
        }

        // Init SRP subwidget
        m_srpForceIsVisible = srPForceIsVisible;
        if (m_srpForceIsVisible) {
            m_srpForceModel = new GPSRPForceModel(composeWidgetId("srpForceModel"),
                    GEnvironment.tr("Solar Radiation Pressure"));
        }

        // Init rediffused SRP subwidget
        m_rediffusedSRPForceIsVisible = rediffusedSRPForceIsVisible;
        if (m_rediffusedSRPForceIsVisible) {
            m_rediffusedSRPForceModel = new GPRediffusedSRPForceModel(
                    composeWidgetId("rediffusedSRPForceModel"),
                    GEnvironment.tr("Rediffused Solar Radiation Pressure"));
        }

        // Init ocean tides subwidget
        m_oceanTidesForceIsVisible = oceanTidesForceIsVisible;
        if (m_oceanTidesForceIsVisible) {
            m_oceanTidesForceModel = new GPOceanTidesForceModel(
                    composeWidgetId("oceanTidesForceModel"), GEnvironment.tr("Ocean Tides"));
        }

        // Init terrestrial tides subwidget
        m_terrestrialTidesForceIsVisible = terrestrialTidesForceIsVisible;
        if (m_terrestrialTidesForceIsVisible) {
            m_terrestrialTidesForceModel = new GPTerrestrialTidesForceModel(
                    composeWidgetId("terrestrialTidesForceModel"),
                    GEnvironment.tr("Terrestrial Tides"));
        }

        // Separators
        for (int i = 0; i < 7; i++) {
            m_separatorsArray[i] = new GSeparator(composeWidgetId("separatorsArray" + i));
        }

        // Setting default constraints
        setDefaultConstraints();

        // Updating widget's status
        updateStatus();

        // Setting which is the class and the rank applying for constraints by type
        activateConstraintsByType(getClass());

        attachCopyPasteMenu();
        
        //JFG MOD
        obsList = new GObserverList();

    }

    /**
     * Method to set the default constraints for the layout.
     */
    @Override
    public final void setDefaultConstraints() {
        setConstraint(0);
    }

    /**
     * Setter for the constraints of the widget.
     * 
     * @param widgetGap left gap from panel border to display the label of the widget.
     */
    public void setConstraint(final int widgetGap) {
        final int GAP = 20;

        final GConstraint labelConstraint = new GConstraint(GConstraint.newline(),
                GConstraint.gapleft(widgetGap), GConstraint.split(2));
        m_label.setInnerDescendantConstraint(labelConstraint, 0);

        // Separators
        for (int i = 0; i < 7; i++) {
            m_separatorsArray[i].setConstraint(new GConstraint(
                    GConstraint.gapleft(widgetGap + GAP), GConstraint.spanx(3),
                    GConstraint.growx(), GConstraint.newline()));
        }

        m_earthPotentialForceModel.setConstraint(widgetGap + GAP);

        if (m_thirdBodyForceModel != null) {
            m_thirdBodyForceModel.setConstraint(widgetGap + GAP);
        }
        if (m_aeroForceModel != null) {
            m_aeroForceModel.setConstraint(widgetGap + GAP);
        }
        if (m_srpForceModel != null) {
            m_srpForceModel.setConstraint(widgetGap + GAP);
        }
        if (m_rediffusedSRPForceModel != null) {
            m_rediffusedSRPForceModel.setConstraint(widgetGap + GAP);

            // This is to avoid horizontal displacement
            // The longest label text : "Absorption Coefficient:"
            final GLabel longestGLabel = m_rediffusedSRPForceModel.getGCheckBoxForceModel()
                    .getGLabelWithIndicator().getGLabel();
            // The always visible label : the widget's label
            final GLabelWithIndicator alwaysVisibleGLabelWithIndicator = m_earthPotentialForceModel
                    .getGLabelForceModel();
            // Avoid horizontal displacement:
            GConstraint
                    .avoidHorizontalDisplacement(longestGLabel, alwaysVisibleGLabelWithIndicator);
        }
        if (m_oceanTidesForceModel != null) {
            m_oceanTidesForceModel.setConstraint(widgetGap + GAP);
        }
        if (m_terrestrialTidesForceModel != null) {
            m_terrestrialTidesForceModel.setConstraint(widgetGap + GAP);
        }
        m_ephemerisTypeChoice.setInnerDescendantConstraint(new GConstraint(GConstraint.newline(),
                GConstraint.split(2), GConstraint.gapleft(widgetGap + 2 * GAP)), 0, 0);

        m_bodyGPOneAxisEllipsoid.setConstraint(widgetGap + 2 * GAP, widgetGap + 3 * GAP);
    }

    /**
     * Set body's shape. If null a {@link GPOneAxisEllipsoid} widget is shown to manually set body's
     * shape.
     * 
     * @param bodyShape body shape to be used to compute force models
     * @throws GPFrameException if the ellipsoid cannot be set in a GPOneAxisEllipsoid widget
     * @throws GArithmeticException if bodyShape eq. radius or flatness are NaN.
     */
    public void setBodyShape(final ExtendedOneAxisEllipsoid bodyShape) throws GPFrameException,
            GArithmeticException {
        if (bodyShape == null) {
            m_bodyGPOneAxisEllipsoidIsVisible = true;
        } else {
            m_bodyGPOneAxisEllipsoid.setEllipsoid(bodyShape);
            m_bodyGPOneAxisEllipsoidIsVisible = false;
            if (m_aeroForceIsVisible) {
                m_aeroForceModel.setBodyShape(bodyShape);
            }
        }
    }

    /**
     * Setter for the assembly to be used for drag force, the SRP and the rediffused SRP force
     * models.
     * 
     * @param assembly the assembly to be set.
     */
    public void setAssembly(final Assembly assembly) {
        if (m_aeroForceIsVisible) {
            m_aeroForceModel.setAssembly(assembly);
        }
        if (m_srpForceIsVisible) {
            m_srpForceModel.setAssembly(assembly);
        }
        if (m_rediffusedSRPForceIsVisible) {
            m_rediffusedSRPForceModel.setAssembly(assembly);
        }
    }

    /**
     * Getter for the assembly used in drag force, SRP and the rediffused SRP force models. Normally
     * it should be the same for the three force models. The priority is drag, then SRP, then
     * rediffused SRP. If one force model is not present, then the assembly from the following one
     * is returned.
     * 
     * @return the used assembly.
     */
    public Assembly getAssembly() {
        Assembly assembly = null;
        if (m_aeroForceIsVisible) {
            assembly = m_aeroForceModel.getAssembly();
        } else if (m_srpForceIsVisible) {
            assembly = m_srpForceModel.getAssembly();
        } else if (m_rediffusedSRPForceIsVisible) {
            assembly = m_rediffusedSRPForceModel.getAssembly();
        }
        return assembly;
    }

    /**
     * Getter for the GLabel of the widget.
     * 
     * @return the GLabel of the widget
     */
    public GLabelWithIndicator getGLabel() {
        return m_label;
    }

    /**
     * Getter for the Earth Potential Force Model subwidget.
     * 
     * @return the Earth Potential Force Model subwidget
     */
    public GPEarthPotentialForceModel getGPEarthPotentialForceModel() {
        // It is allways visible
        final GPEarthPotentialForceModel model = m_earthPotentialForceModel;
        return model;
    }

    /**
     * Getter for the Third body Force Model subwidget.
     * 
     * @return the Third body Force Model subwidget
     */
    public GPThirdBodyForceModel getGPThirdBodyForceModel() {
        GPThirdBodyForceModel model = null;
        if (m_thirdBodyForceIsVisible) {
            model = m_thirdBodyForceModel;
        }
        return model;
    }

    /**
     * Getter for the Atmospheric Force Model subwidget.
     * 
     * @return the Atmospheric Force Model subwidget
     */
    public GPAeroForceModel getGPAeroForceModel() {
        GPAeroForceModel model = null;
        if (m_aeroForceIsVisible) {
            model = m_aeroForceModel;
        }
        return model;
    }

    /**
     * Getter for the Solar Radiation Pressure Force Model subwidget.
     * 
     * @return the Solar Radiation Pressure Force Model subwidget
     */
    public GPSRPForceModel getGPSRPForceModel() {
        GPSRPForceModel model = null;
        if (m_srpForceIsVisible) {
            model = m_srpForceModel;
        }
        return model;
    }

    /**
     * Getter for the Rediffused Solar Radiation Pressure Force Model subwidget.
     * 
     * @return the Rediffused Solar Radiation Pressure Force Model subwidget
     */
    public GPRediffusedSRPForceModel getGPRediffusedSRPForceModel() {
        GPRediffusedSRPForceModel model = null;
        if (m_rediffusedSRPForceIsVisible) {
            model = m_rediffusedSRPForceModel;
        }
        return model;
    }

    /**
     * Getter for the Ocean Tides Force Model subwidget.
     * 
     * @return the Ocean Tides Force Model subwidget
     */
    public GPOceanTidesForceModel getGPOceanTidesForceModel() {
        GPOceanTidesForceModel model = null;
        if (m_oceanTidesForceIsVisible) {
            model = m_oceanTidesForceModel;
        }
        return model;
    }

    /**
     * Getter for the Terrestrial Tides Force Model subwidget.
     * 
     * @return the Terrestrial Tides Force Model subwidget
     */
    public GPTerrestrialTidesForceModel getGPTerrestrialTidesForceModel() {
        GPTerrestrialTidesForceModel model = null;
        if (m_terrestrialTidesForceIsVisible) {
            model = m_terrestrialTidesForceModel;
        }
        return model;
    }

    /**
     * Getter for the GChoice defining the ephemeris type.
     * 
     * @return the GChoice defining the ephemeris type
     */
    public GChoice getGChoiceEphemerisType() {
        return m_ephemerisTypeChoice;
    }

    /**
     * Getter for the GPOneAxisEllipsoid defining the Earth's ellipsoid.
     * 
     * @return the GPOneAxisEllipsoid defining the Earth's ellipsoid
     */
    public GPOneAxisEllipsoid getGPOneAxisEllipsoidEarthShape() {
        return m_bodyGPOneAxisEllipsoid;
    }

    /**
     * Getter for the array of GSeparators in the widget.
     * 
     * @return the array of GSeparators in the widget
     */
    public GSeparator[] getGSeparatorsArray() {
        return m_separatorsArray;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setEnabled(final boolean b) {
        m_label.setEnabled(b);
        m_earthPotentialForceModel.setEnabled(b);
        m_thirdBodyForceModel.setEnabled(b);
        m_aeroForceModel.setEnabled(b);
        m_srpForceModel.setEnabled(b);
        m_rediffusedSRPForceModel.setEnabled(b);
        m_oceanTidesForceModel.setEnabled(b);
        m_terrestrialTidesForceModel.setEnabled(b);
        m_bodyGPOneAxisEllipsoid.setEnabled(b);
        m_ephemerisTypeChoice.setEnabled(b);
        // If we ask to enable the orbit but it is in error status not all should be enabled.
        if (b) {
            // updateIsSelectionEnabled();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void attachCopyPasteMenu(final String label, final CopyPasteOperation... operationList) {
        m_label.attachCopyPasteMenu(label, operationList);
    }

    /**
     * Updates the status of the widget.
     */
    private void updateStatus() {

        GStatus newStatus = null;
        String newToolTip = null;

        final GCondensedStatus condensedStatus = new GCondensedStatus(this);
        newStatus = condensedStatus.getStatus();

        // Setting tooltip message indicating which component is in error status
        final String COMMA = ", ";
        switch (newStatus) {
            case OK:
                // Removing label tooltip text
                newToolTip = null;
                break;

            case WARN:

                // Setting warning components list
                final StringBuffer warningBuffer = new StringBuffer();
                boolean firstTime = true;
                for (final GComponent component : condensedStatus.getWarningComponentList()) {
                    if (firstTime) {
                        // Avoid to put a comma in first place
                        firstTime = false;
                    } else {
                        warningBuffer.append(COMMA);
                    }
                    warningBuffer.append(component.getNameInConfigFile());
                }
                final String warningList = warningBuffer.toString();

                // List of components in warning status
                newToolTip = GEnvironment.getWarningTooltipMessage(warningList);
                // Log event
                GEnvironment.getGLog().writeLogWarn(Type.OPER, getWidgetId(),
                        GEnvironment.tr("Value in warning range."));
                break;

            case ERROR:

                // Setting error components list
                final StringBuffer errorBuffer = new StringBuffer();
                firstTime = true;
                for (final GComponent component : condensedStatus.getErrorComponentList()) {
                    if (firstTime) {
                        // Avoid to put a comma in first place
                        firstTime = false;
                    } else {
                        errorBuffer.append(COMMA);
                    }
                    errorBuffer.append(component.getNameInConfigFile());
                }
                final String errorList = errorBuffer.toString();

                // List of components in error status
                newToolTip = GEnvironment.getErrorTooltipMessage(errorList);
                // Log event
                GEnvironment.getGLog().writeLogWarn(Type.OPER, getWidgetId(),
                        GEnvironment.tr("Value in error range."));
                break;
        }
        if (newStatus != getStatus()) {
            setStatus(newStatus);
            m_label.setStatus(newStatus);
            m_label.setToolTipText(newToolTip);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void updateIsModifiedIndicator() {
        m_label.setIndicatorVisible(getDisplayIsModifiedIndicator());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean isEnabled() {
        return m_label.isEnabled();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void generic() throws GException {

        // Earth potential
        put(m_earthPotentialForceModel);
        put(m_separatorsArray[0]);

        // Third Body
        if (m_thirdBodyForceIsVisible) {
            put(m_thirdBodyForceModel);
            put(m_separatorsArray[1]);
        }

        // Aero
        if (m_aeroForceIsVisible) {
            put(m_aeroForceModel);
            put(m_separatorsArray[2]);
        }

        // SRP
        if (m_srpForceIsVisible) {
            put(m_srpForceModel);
            put(m_separatorsArray[3]);
        }

        // Redifussed SRP
        if (m_rediffusedSRPForceIsVisible) {
            put(m_rediffusedSRPForceModel);
            put(m_separatorsArray[4]);
        }

        // Ocean tides
        if (m_oceanTidesForceIsVisible) {
            put(m_oceanTidesForceModel);
            put(m_separatorsArray[5]);
        }

        // Terrestrial tides
        if (m_terrestrialTidesForceIsVisible) {
            put(m_terrestrialTidesForceModel);
            put(m_separatorsArray[6]);
        }

        // If sun or moon are requiered
        if (isEphemerisTypeVisible()) {
            put(m_ephemerisTypeChoice);
        }

        // If Earth's shape is requiered
        if (m_bodyGPOneAxisEllipsoidIsVisible) {
            put(m_bodyGPOneAxisEllipsoid);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void display() throws GException {
        put(m_label);
        generic();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void read() throws GException {
        generic();

        // Updating status and modification indicator
        updateStatus();
        updateIsModifiedIndicator();
        notifyObservers();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void write() throws GException {
        generic();

        // Updating status and modification indicator
        updateIsModifiedIndicator();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void clear() throws GException {
        // Earth potential
        put(m_earthPotentialForceModel);

        // Third Body
        if (m_thirdBodyForceIsVisible) {
            put(m_thirdBodyForceModel);
        }

        // Aero
        if (m_aeroForceIsVisible) {
            put(m_aeroForceModel);
        }

        // SRP
        if (m_srpForceIsVisible) {
            put(m_srpForceModel);
        }

        // Redifussed SRP
        if (m_rediffusedSRPForceIsVisible) {
            put(m_rediffusedSRPForceModel);
        }

        // Ocean tides
        if (m_oceanTidesForceIsVisible) {
            put(m_oceanTidesForceModel);
        }

        // Terrestrial tides
        if (m_terrestrialTidesForceIsVisible) {
            put(m_terrestrialTidesForceModel);
        }

        // Ephemeris type (it allwais exists, even if it not visible the clear must be done).
        put(m_ephemerisTypeChoice);

        // Earth's shape (it allwais exists, even if it not visible the clear must be done).
        put(m_bodyGPOneAxisEllipsoid);
        m_bodyGPOneAxisEllipsoidIsVisible = m_earthGPOneAxisEllipsoidIsVisibleByDefault;

        // Updating status and modification indicator
        updateStatus();
        updateIsModifiedIndicator();
    }

    /**
     * Get selected ephemeris type as a {@link TypeEphemeris} enum value.
     * 
     * @return the selected ephemeris type
     */
    public TypeEphemeris getSelectedEphemerisType() {
        return (TypeEphemeris) m_ephemerisTypeChoice.getValueEnum();
    }

    /**
     * {@inheritDoc}
     * 
     * @throws GPForcesException if the assembly has not been initialized or if some internal error
     *         occurs
     */
    @Override
    public ForceModelsData getPatriusObject() throws GPForcesException {

        // Getting Earth's ellipsoid
        final ExtendedOneAxisEllipsoid earthShape = getBodyShape();

        // Equatorial radius
        final double eqRadius = earthShape.getEquatorialRadius();
        // Frame
        final Frame earthFrame = earthShape.getBodyFrame();

        // Celestial bodies for the Sun and the Moon
        final TypeEphemeris ephemerisType = getSelectedEphemerisType();
        final CelestialBody sun = getSun();
        final CelestialBody moon = buildMoon(ephemerisType, eqRadius);

        // 1- Earth Potential force model
        m_earthPotentialForceModel.setEarthFrame(earthFrame);
        final ForceModel earthPotentialForceModel = m_earthPotentialForceModel.getPatriusObject();
        final double mu = m_earthPotentialForceModel.getMu();

        ThirdBodyAttraction moonForceModel = null;
        ThirdBodyAttraction sunForceModel = null;
        ThirdBodyAttraction venusForceModel = null;
        ThirdBodyAttraction marsForceModel = null;
        ThirdBodyAttraction jupiterForceModel = null;
        if (m_thirdBodyForceIsVisible) {
            // 2- Moon perturbation force model, null if not selected.
            m_thirdBodyForceModel.setMoon(moon);
            moonForceModel = m_thirdBodyForceModel.getMoonForceModel();

            // 3- Sun perturbation force model, null if not selected
            m_thirdBodyForceModel.setSun(sun);
            sunForceModel = m_thirdBodyForceModel.getSunForceModel();

            // 4- Venus perturbation force model, null if not selected
            venusForceModel = m_thirdBodyForceModel.getVenusForceModel();

            // 5- Mars perturbation force model, null if not selected
            marsForceModel = m_thirdBodyForceModel.getMarsForceModel();

            // 6- Jupiter perturbation force model, null if not selected
            jupiterForceModel = m_thirdBodyForceModel.getJupiterForceModel();
        }

        // 7- Aero force model
        DragForce dragForce = null;
        if (m_aeroForceIsVisible) {
            try {
                m_aeroForceModel.setBodyShape(earthShape);
            } catch (final GArithmeticException e) {
                // If earth shape contains NaN (impossible comming from widget)
                throw new GRuntimeException(e);
            }
            m_aeroForceModel.setSun(sun);
            dragForce = m_aeroForceModel.getPatriusObject();
        }

        // 8- Add SRP force model
        SolarRadiationPressureEllipsoid solarRadiationPressure = null;
        if (m_srpForceIsVisible) {
            try {
                m_srpForceModel.setEarthBodyShape(earthShape);
            } catch (final GArithmeticException e) {
                // If earth shape contains NaN (impossible comming from widget)
                throw new GRuntimeException(e);
            }
            m_srpForceModel.setSun(sun);
            solarRadiationPressure = m_srpForceModel.getPatriusObject();
        }

        // 9- Add Rediffused SRP force model
        RediffusedRadiationPressure rediffusedSRP = null;
        if (m_rediffusedSRPForceIsVisible) {
            m_rediffusedSRPForceModel.setEarthFrame(earthFrame);
            m_rediffusedSRPForceModel.setSun(sun);
            rediffusedSRP = m_rediffusedSRPForceModel.getPatriusObject();
        }

        // 10- Add ocean tides force model
        OceanTides oceanTidesForceModel = null;
        if (m_oceanTidesForceIsVisible) {
            try {
                m_oceanTidesForceModel.setMu(mu);
                m_oceanTidesForceModel.setEqRadius(eqRadius);
            } catch (final GArithmeticException e) {
                // if mu from GPEarthPotentialForceModel is NaN (unusual) or if eqRadius is NaN
                // (impossible
                // comming fron a GEntryReal)
                throw new GRuntimeException(e);
            }
            m_oceanTidesForceModel.setEarthFrame(earthFrame);
            oceanTidesForceModel = m_oceanTidesForceModel.getPatriusObject();
        }

        // 11- Add terrestrial tides force model
        TerrestrialTides terrestrialTidesForceModel = null;
        if (m_terrestrialTidesForceIsVisible) {
            try {
                m_terrestrialTidesForceModel.setMu(mu);
                m_terrestrialTidesForceModel.setEqRadius(eqRadius);
            } catch (final GArithmeticException e) {
                // eqRadius comes from a widget. mu comes from reading a data file and should not be
                // NaN.
                throw new GRuntimeException(e);
            }
            m_terrestrialTidesForceModel.setEarthFrame(earthFrame);
            terrestrialTidesForceModel = m_terrestrialTidesForceModel.getPatriusObject();
        }

        return new ForceModelsData(earthPotentialForceModel, null, solarRadiationPressure,
                rediffusedSRP, moonForceModel, sunForceModel, venusForceModel, marsForceModel,
                jupiterForceModel, oceanTidesForceModel, terrestrialTidesForceModel, dragForce,
                null, null, null);
    }

    /**
     * {@inheritDoc}
     * 
     * @param args The parameters are 0: Assembly or 0: ExtendedOneAxisEllipsoid, 1: Assembly.
     */
    @Override
    public ForceModelsData getPatriusObject(final Object... args) throws GException {
        final boolean bodyShapeVisible = m_bodyGPOneAxisEllipsoidIsVisible;
        ForceModelsData patriusObject = null;
        // The body shape is taken from the widget
        if (args.length == 1) {

            // Check the number of input arguments and their classes
            new ArgsChecker(args, Assembly.class);

            final Assembly prevAssembly = getAssembly();
            setAssembly((Assembly) args[0]);

            patriusObject = getPatriusObject();

            setAssembly(prevAssembly);

        } else {

            // Check the number of input arguments and their classes
            new ArgsChecker(args, ExtendedOneAxisEllipsoid.class, Assembly.class);

            final ExtendedOneAxisEllipsoid prevBodyShape = getBodyShape();
            setBodyShape((ExtendedOneAxisEllipsoid) args[0]);

            final Assembly prevAssembly = getAssembly();
            setAssembly((Assembly) args[1]);

            patriusObject = getPatriusObject();

            setBodyShape(prevBodyShape);
            setAssembly(prevAssembly);
        }
        setBodyShapeWidgetsVisible(bodyShapeVisible);
        return patriusObject;
    }

    /**
     * Builts the Moon celestial body depending on the ephemeris type.
     * 
     * @param ephemerisType type of ephemeris to use to create the Moon celestial body
     * @param eqRadius Earth's equatorial radius (requiered for MEEUS Stela ephemeris type)
     * @return the Moon celestial body
     * @throws GPForcesException if the celestial body cannot be built
     */
    private static CelestialBody buildMoon(final TypeEphemeris ephemerisType, final double eqRadius)
            throws GPForcesException {
        CelestialBody moon = null;
        try {
            switch (ephemerisType) {
                case JPL:
                    moon = CelestialBodyFactory.getMoon();
                    break;

                case MEEUS:
                    moon = new MeeusMoon();
                    break;

                case MEEUSSTELA:
                    moon = new MeeusMoonStela(eqRadius);
                    break;
            }
        } catch (final PatriusException e) {
            throw new GPForcesException(e);
        } catch (final ArithmeticException e) {
            // if moon body cannot be built.
            throw new GRuntimeException(e);
        }
        return moon;
    }

    /**
     * Getter for the Sun celestial body to be used for drag force model creation. Requiered for the
     * atmosphere model.
     * 
     * @return the internal Sun celestial body.
     */
    public final CelestialBody getSun() {
        final CelestialBody sun = m_sunBodiesMap.get(getSelectedEphemerisType());
        return sun;
    }

    /**
     * Get Aero force model multiplicative factor.
     * 
     * @return the multiplicative factor for Aero force model
     */
    public double getMultiplicativeFactorAero() {
        return m_aeroForceModel.getMultiplicativeFactor();
    }

    /**
     * Get SRP force model multiplicative factor.
     * 
     * @return the multiplicative factor for SRP force model
     */
    public double getMultiplicativeFactorSRP() {
        return m_srpForceModel.getMultiplicativeFactor();
    }

    /**
     * Get redifused SRP force model albedo multiplicative factor.
     * 
     * @return the albedo multiplicative factor for rediffused SRP force model
     */
    public double getMultiplicativeFactorRediffusedSRPAlbedo() {
        return m_rediffusedSRPForceModel.getMultiplicativeFactorAlbedo();
    }

    /**
     * Get redifused SRP force model infrared multiplicative factor.
     * 
     * @return the infrared multiplicative factor for rediffused SRP force model
     */
    public double getMultiplicativeFactorRediffusedSRPInfrared() {
        return m_rediffusedSRPForceModel.getMultiplicativeFactorInfrared();
    }

    /**
     * Boolean to know if the third body force model is active.
     * 
     * @return true if the third body force model is present in the panel and it is selected.
     */
    public boolean isThirdBodyForceModelActive() {
        return m_thirdBodyForceIsVisible && m_thirdBodyForceModel.isSelected();
    }

    /**
     * Boolean to know if the aero force model is active.
     * 
     * @return true if the aero force model is present in the panel and it is selected
     */
    public boolean isAeroForceModelActive() {
        return m_aeroForceIsVisible && m_aeroForceModel.isSelected();
    }

    /**
     * Boolean to know if the SRP force model is active.
     * 
     * @return true if the SRP force model is present in the panel and it is selected
     */
    public boolean isSRPForceModelActive() {
        return m_srpForceIsVisible && m_srpForceModel.isSelected();
    }

    /**
     * Boolean to know if the rediffused SRP force model is active.
     * 
     * @return true if the rediffused SRP force model is present in the panel and it is selected
     */
    public boolean isRediffusedSRPForceModelActive() {
        return m_rediffusedSRPForceIsVisible && m_rediffusedSRPForceModel.isSelected();
    }

    /**
     * Boolean to know if the ocean tides force model is active.
     * 
     * @return true if the ocean tides force model is present in the panel and it is selected
     */
    public boolean isOceanTidesForceModelActive() {
        return m_oceanTidesForceIsVisible && m_oceanTidesForceModel.isSelected();
    }

    /**
     * Boolean to know if the terrestrial tides force model is active.
     * 
     * @return true if the terrestrial tides force model is present in the panel and it is selected
     */
    public boolean isTerrestrialTidesForceModelActive() {
        return m_terrestrialTidesForceIsVisible && m_terrestrialTidesForceModel.isSelected();
    }

    /**
     * Boolean to know if the ephemeris type choice is visible.
     * 
     * @return true if the ephemeris type choice is present in the panel
     */
    public boolean isEphemerisTypeVisible() {
        // If sun or moon third body forces
        final boolean boolean1 = m_thirdBodyForceIsVisible
                && (m_thirdBodyForceModel.isMoonSelected() || m_thirdBodyForceModel.isSunSelected());
        // If DTM or MSISE200 aero forces are selected
        final boolean boolean2 = (m_aeroForceIsVisible && m_aeroForceModel.isSelected())
                && (m_aeroForceModel.getSelectedAtmosphericModel() == AtmosphereType.DTM2000 || m_aeroForceModel
                        .getSelectedAtmosphericModel() == AtmosphereType.MSISE2000);
        // If srp or rediffused are selected
        final boolean boolean3 = (m_srpForceIsVisible && m_srpForceModel.isSelected())
                || (m_rediffusedSRPForceIsVisible && m_rediffusedSRPForceModel.isSelected());

        return boolean1 || boolean2 || boolean3;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void updateCondensedStatus(final GCondensedStatus condensedStatus) {
        condensedStatus.update(m_earthPotentialForceModel, m_bodyGPOneAxisEllipsoid);
        if (isThirdBodyForceModelActive()) {
            condensedStatus.update(m_thirdBodyForceModel);
        }
        if (isAeroForceModelActive()) {
            condensedStatus.update(m_aeroForceModel);
        }
        if (isSRPForceModelActive()) {
            condensedStatus.update(m_srpForceModel);
        }
        if (isRediffusedSRPForceModelActive()) {
            condensedStatus.update(m_rediffusedSRPForceModel);
        }
        if (isOceanTidesForceModelActive()) {
            condensedStatus.update(m_oceanTidesForceModel);
        }
        if (isTerrestrialTidesForceModelActive()) {
            condensedStatus.update(m_terrestrialTidesForceModel);
        }
        if (isEphemerisTypeVisible()) {
            condensedStatus.update(m_ephemerisTypeChoice);
        }
    }

    /**
     * Getter for the body's ellipsoid that defines its shape.
     * 
     * @return the body's ellipsoid
     * @throws GPForcesException if EOP data is not initialized
     */
    public final ExtendedOneAxisEllipsoid getBodyShape() throws GPForcesException {
        try {
            return m_bodyGPOneAxisEllipsoid.getPatriusObject();
        } catch (final GPFrameException e) {
            throw new GPForcesException(e);
        }
    }

    /**
     * Method to force the GPOneAxisEllipsoid widget to be visible/invisible.
     * 
     * @param b if true the widget will be visible, if false it will not.
     */
    public void setBodyShapeWidgetsVisible(final boolean b) {
        m_bodyGPOneAxisEllipsoidIsVisible = b;
    }

    /**
     * Get the atmospheric model from widget data.
     * 
     * @return the atmospheric model.
     * @throws GPForcesException if there is a problem with solar activity provider
     */
    public Atmosphere getAtmosphere() throws GPForcesException {
        Atmosphere atm = null;
        final GPAeroForceModel aeroForceModel = getGPAeroForceModel();
        if (aeroForceModel != null) {
            atm = aeroForceModel.getAtmosphere();
        }
        return atm;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public GComponent getSubComponent(final int position) throws GComponentException {
        GComponent subcomponent = null;
        switch (position) {
            case 0:
                subcomponent = m_label;
                break;
            case 1:
                subcomponent = m_earthPotentialForceModel;
                break;
            case 2:
                subcomponent = m_thirdBodyForceModel;
                break;
            case 3:
                subcomponent = m_aeroForceModel;
                break;
            case 4:
                subcomponent = m_srpForceModel;
                break;
            case 5:
                subcomponent = m_rediffusedSRPForceModel;
                break;
            case 6:
                subcomponent = m_oceanTidesForceModel;
                break;
            case 7:
                subcomponent = m_terrestrialTidesForceModel;
                break;
            case 8:
                subcomponent = m_ephemerisTypeChoice;
                break;
            case 9:
                subcomponent = m_bodyGPOneAxisEllipsoid;
                break;
            case 10:
                // BEGIN_EMBEDDED_NUMBER
            case 11:
            case 12:
            case 13:
            case 14:
            case 15:
            case 16:
                // END_EMBEDDED_NUMBER
                if (m_separatorsArray != null) {
                    subcomponent = m_separatorsArray[position - 10];
                }
                break;
            default:
                launchInvalidPositionException(position);
                break;
        }
        return subcomponent;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean isModified() {
        final GCondensedIsModified condensedStatus = new GCondensedIsModified(this);
        final boolean modified = condensedStatus.getIsModified();
        return modified;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    @SuppressWarnings("PMD.UnusedFormalParameter")
    public void before(final GEvent e) throws GException {
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void after(final GEvent e) throws GException {
        // Handling copy-paste event
        boolean handled;
        try {
            handled = handleCopyPasteEvent(m_label, e);
        } catch (final GFileManipulatorException exception) {
            throw new GPEventException(exception);
        }

        // If it is a copy-paste event, exit after handling.
        if (handled) {
            return;
        }

        // Due to the fact that the method getAtmosphere in GPAeroForceModel needs to know the
        // bodyShape and
        // the sun, they must be kept updated at any time:
        if (e.contains(m_bodyGPOneAxisEllipsoid) && m_aeroForceIsVisible) {
            m_aeroForceModel.setBodyShape(getBodyShape());
        }
        if (e.contains(m_ephemerisTypeChoice) && m_aeroForceIsVisible) {
            m_aeroForceModel.setSun(getSun());
        }

        // Notify GPVehicle if some forces have been modified
        if (e.contains(m_aeroForceModel) || e.contains(m_srpForceModel) || e.contains(m_rediffusedSRPForceModel)) {
            notifyObservers();
        }
        
        updateStatus();
        updateIsModifiedIndicator();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void updateIsModified(final GCondensedIsModified condensedIsModified) {
        condensedIsModified.update(m_earthPotentialForceModel, m_bodyGPOneAxisEllipsoid);
        if (isThirdBodyForceModelActive()) {
            condensedIsModified.update(m_thirdBodyForceModel);
        }
        if (isAeroForceModelActive()) {
            condensedIsModified.update(m_aeroForceModel);
        }
        if (isSRPForceModelActive()) {
            condensedIsModified.update(m_srpForceModel);
        }
        if (isRediffusedSRPForceModelActive()) {
            condensedIsModified.update(m_rediffusedSRPForceModel);
        }
        if (isOceanTidesForceModelActive()) {
            condensedIsModified.update(m_oceanTidesForceModel);
        }
        if (isTerrestrialTidesForceModelActive()) {
            condensedIsModified.update(m_terrestrialTidesForceModel);
        }
        if (isEphemerisTypeVisible()) {
            condensedIsModified.update(m_ephemerisTypeChoice);
        }
    }

    //JFG DEB MOD
    /** {@inheritDoc} */
    @Override
    public void registerObserver(final GObserver observer) {
        obsList.registerObserver(observer);
//        if (observer instanceof GPVehicle) {
//            observerGPVehicle = observer;
//        }
    }

    /** {@inheritDoc} */
    @Override
    public void unregisterObserver(final GObserver observer) {
        // Nothing to do
    }

    /** {@inheritDoc} */
    @Override
    public void notifyObservers(final Object... args) {
        obsList.notifyObservers(this, args);
//        if (observerGPVehicle != null) {
//            observerGPVehicle.notify(this, args);
//        }
    }
    //JFG END MOD
}
